#include "__whiledir.h"
#include "Exception.h"
#include "__mysql.h"
#include <iostream>
#include <filesystem>
#include <vector>
#include <algorithm>

namespace jet {

   __whiledir::__whiledir(coreutils::ZString &in, coreutils::MString &parentOut, Global &global, Tag *parent, Tag *local) : Tag(in, parentOut, global, parent, this) {
      if(!keywordDefined("path"))
	throw coreutils::Exception("whiledir tag must specify a path.");
      resolveKeyword("path");
      resolveKeyword("sort");
      if(keywordDefined("sort") && (keywords["sort"] == "true")) {
	 std::vector<std::filesystem::directory_entry> entries;
	 for(auto const &entry : std::filesystem::directory_iterator(variables["path"].str())) 
	   entries.push_back(entry);
	 std::sort(entries.begin(), entries.end(), [](const auto &a, const auto &b) { return a.path() < b.path(); });
	 for(const auto &entry : entries) {
	    if(keywordDefined("fullpath")) {		 
	       resolveKeyword("fullpath");
	       global.variables[keywords["fullpath"]] = entry.path();
	    }	    
	    if(keywordDefined("filename")) {
	       resolveKeyword("filename");
	       global.variables[keywords["filename"]] = entry.path().filename();
	    }	    
	    if(keywordDefined("filenamenoextension")) {
	       resolveKeyword("filenamenoextension");
	       global.variables[keywords["filenamenoextension"]] = entry.path().stem();
	    }	    
	    processContainer(container);
	    container.reset();	    
	 }	 
      } else {	 
	 for(auto const &entry : std::filesystem::directory_iterator(variables["path"].str())) {
	    if(keywordDefined("fullpath")) {
		resolveKeyword("fullpath"); 
	       global.variables[keywords["fullpath"]] = entry.path();
	    }	    
	    if(keywordDefined("filename")) {
	       resolveKeyword("filename");
	       global.variables[keywords["filename"]] = entry.path().filename();
	    }	    
	    if(keywordDefined("filenamenoextension")) {
	       resolveKeyword("filenamenoextension");
	       global.variables[keywords["filenamenoextension"]] = entry.path().stem();
	    }	    
	    processContainer(container);
	    container.reset();
	 }
      }
      
   }
   
}
